package require SOAP

global SOAPServerHost SOAPServerPort

# -------------------------------------------------------------------------

# Description:
#   Procedure to generate the XML data for a configured SOAP procedure.
#   This is the XDAQ -wrapProc procedure
# Parameters:
#   procVarName - the path of the SOAP method configuration variable
#   args        - the arguments for this SOAP method
# Result:
#   XML data containing the SOAP method call.
# Notes:
#   We permit a small number of option to be specified on the method call
#   itself. -headers is used to set SOAP Header elements and -attr can be
#   used to set additional XML attributes on the method element (needed for
#   UDDI.)
#

proc ::SOAP::XDAQ_Make_Request {procVarName args}  {
   upvar $procVarName procvar
   insert "$procVarName"
   insert "$args"

    set procName [lindex [split $procVarName {_}] end]
    set params  $procvar(params)
    set name    $procvar(name)
    set uri     $procvar(uri)
    set soapenv $procvar(version)
    set soapenc $procvar(encoding)

    # Check for options (ie: -header) give up on the fist non-matching arg.
    array set opts {-headers {} -attributes {}}
    while {[string match -* [lindex $args 0]]} {
        switch -glob -- [lindex $args 0] {
            -header* {
                set opts(-headers) [concat $opts(-headers) [lindex $args 1]]
                set args [lreplace $args 0 0]
            }
            -attr* {
                set opts(-attributes) [concat $opts(-attributes) [lindex $args 1]]
                set args [lreplace $args 0 0]
            }
            -- {
                set args [lreplace $args 0 0]
                break
            }
            default {
                # stop option processing at the first invalid option.
                break
            }
        }
        set args [lreplace $args 0 0]
    }

    # check for variable number of params and set the num required.
    if {[lindex $params end] == "args"} {
        set n_params [expr ( [llength $params] - 1 ) / 2]
    } else {
        set n_params [expr [llength $params] / 2]
    }

    # check we have the correct number of parameters supplied.
    if {[llength $args] < $n_params} {
        set msg "wrong # args: should be \"$procName"
        foreach { id type } $params {
            append msg " " $id
        }
        append msg "\""
        error $msg
    }

    set doc [dom::DOMImplementation create]
    set envx [dom::document createElement $doc "SOAP-ENV:Envelope"]

    dom::element setAttribute $envx "xmlns:SOAP-ENV" $soapenv

    # Insert the empty Header
    set headelt [dom::document createElement $envx "SOAP-ENV:Header"]

    # Insert the body element and atributes.
    set bod [dom::document createElement $envx "SOAP-ENV:Body"]

    set cmd [dom::document createElement $bod "$name" ]

    # Insert any method attributes
    if {$opts(-attributes) != {}} {
        foreach {atname atvalue} $opts(-attributes) {
            dom::element setAttribute $cmd $atname $atvalue
        }
    }

    # Insert the parameter header
    set ParamH [dom::document createElement $cmd "Parameter"]
    dom::element setAttribute $ParamH name parameters
    dom::element setAttribute $ParamH type bag

    # insert the parameters.
    set param_no 0
    foreach {key type} $params {
        set val [lindex $args $param_no]
        set Param [dom::document createElement $ParamH "Parameter"]
        dom::element setAttribute $Param name $key
        dom::element setAttribute $Param type $type
        dom::document createTextNode $Param $val
        incr param_no
    }

    # We have to strip out the DOCTYPE element though. It would be better to
    # remove the DOM node for this, but that didn't work.
    set prereq [dom::DOMImplementation serialize $doc]
    set req {}
    dom::DOMImplementation destroy $doc              ;# clean up
    regsub "<!DOCTYPE\[^>\]*>\r?\n?" $prereq {} req  ;# hack
    set req [encoding convertto utf-8 $req]          ;# make it UTF-8
    return $req                                      ;# return the XML data
}



SOAP::create random \
   -uri  "urn:tclsoap:DomainTest" \
   -proxy http://$SOAPServerHost:$SOAPServerPort/domaintest  \
   -params {num float}

#SOAP::configure random -transport SOAP::Transport::print::print -wrapProc ::SOAP::XDAQ_Make_Request

#SOAP::configure random  -wrapProc ::SOAP::XDAQ_Make_Request

